package nl.uvt.locator;

import java.io.*;
import java.net.*; 
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import nu.xom.*;

import org.apache.log4j.*;


public class Tester extends HttpServlet {

	public static Properties testerConfig = null;
	public static Properties httpConfig = null;
	public static final String xhtmlNamespace  = "http://www.w3.org/1999/xhtml";

	static XPathContext xhtmlContext = new XPathContext();

	static private String []my_prefixPaths;

	private void fillContext(){
		xhtmlContext.addNamespace("xhtml", xhtmlNamespace);
	}

	static public Logger logger;

	private File getFile( String[] prefixes, String name ){
		try {
			int i=0;
			while ( i < prefixes.length ){
				String fullName = prefixes[i] + File.separator + name;
				File file = new File( fullName );
				if ( file.exists() ){
					return file;
				}
				++i;
			}
		}
		catch ( Exception e ){
		}
		return null;
	}

	private File getFile( String name ){
		return getFile( my_prefixPaths, name );
	}

	private Properties readConfig( String name ){
		Properties result = null;
		try {
			logger.debug("get file:" + name );
			File configFile = getFile( name );
			if ( configFile != null ){
				FileInputStream is = new FileInputStream(configFile);
				result = new Properties();
				result.load( is );
			}		
		}
		catch ( Exception e ){
			logger.error("readConfig problem: "+e.getMessage() );
		}
		return result;
	}

	public void init() throws ServletException{
		String prefixPath = getServletContext().getInitParameter("prefixPath" );
		my_prefixPaths = prefixPath.split(":");
		if ( my_prefixPaths.length == 0 ){
			throw new ServletException("no prefix paths for configuration found" );
		}
		String configName = null;
		File l4j = getFile( "log4j.properties");
		if ( l4j != null){
			configName = l4j.getAbsolutePath();
			PropertyConfigurator.configure(configName);
		}
		else {
			BasicConfigurator.configure();
		}
		logger = Logger.getLogger(Tester.class);
		if ( configName != null )
			logger.debug("log4j configured from " + configName );
		else
			logger.debug("couldn't get log4j.properties. logging to System.out");
		fillContext();
	}

	public void doGet( HttpServletRequest request, 
			HttpServletResponse response ) throws ServletException, IOException {

		response.setContentType("text/html");
		response.setCharacterEncoding("UTF-8");
		response.setHeader("Cache-Control", "max-age=0, no-cache, must-revalidate");
		OutputStream os = response.getOutputStream();
		PrintStream out = new PrintStream( os, true );
		String fileName = "tester.cfg";
		testerConfig = readConfig( fileName );
		if ( testerConfig == null ){
			out.println("unable to read configfile " + fileName );
			logger.error("reading configfile '" + fileName + "' failed");
			return;
		}
		logger.debug("config:" + testerConfig.toString());
		String httpSettings = testerConfig.getProperty("httpDefaults");
		if ( httpSettings != null ){
			httpConfig = readConfig(httpSettings);
			if ( httpConfig == null ){
				logger.warn("reading http configfile '" + httpSettings  + "' failed. (using defaults)");
			}
			else
				logger.debug("HTTP config:" + httpConfig.toString());
		}

		out.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
		out.println("<html>");
		out.println("<head>");
		out.println( "<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">" );
		out.println("<title>Tester output</title>");
		out.println("</head>");
		out.println("<body>");
		String method = request.getParameter( "method" );
		if ( method != null && !method.equals("")){
			doTestLinks( out, request );
		}
		else {
			doTestDb( out, request );
		}
	}

	void doTestLinks( PrintStream out, HttpServletRequest request ){
		String method = request.getParameter( "method" );
		String select = request.getParameter( "select" );
		Document testDoc = getTestDoc( testerConfig.getProperty("testLinks"));
		if ( testDoc != null ){
//			logger.debug("found testdoc:");
//			logger.debug(testDoc.toXML());
			Nodes nodes = null;
			if ( !method.equals("all") ){
				String query = makeQuery( "method", method, select );
				logger.debug("Xpath query="+query);
				nodes = testDoc.query( query, xhtmlContext );
				if ( nodes.size() == 0 ){
					out.println( "no tests for method " + method + " found." );
					return;
				}
				logger.debug("found " + nodes.size() + " tests for " + method );
			}
			else {
				String query = makeQuery( select );
				nodes = testDoc.query( query, xhtmlContext );
				if ( nodes.size() == 0 ){
					out.println( "no tests found." );
					return;
				}
				logger.debug("found " + nodes.size() + " tests" );
			}
			Integer startIndex = 1;
			Integer endIndex = nodes.size();
			out.println("Testing links ");
			out.println("<table border=\"1\">");
			for( int i=startIndex-1; i < endIndex; ++i){
				String ln = nodes.get(i).query("xhtml:method", xhtmlContext).get(0).getValue().toString();
				String db = nodes.get(i).query("xhtml:database", xhtmlContext).get(0).getValue().toString();
				String query = nodes.get(i).query("xhtml:query",xhtmlContext).get(0).getValue().toString();
				String pos = "1";
				Nodes posNode = nodes.get(i).query("xhtml:position",xhtmlContext);
				if ( posNode.size() > 0 )
					pos = posNode.get(0).getValue().toString();
				Nodes variants = nodes.get(i).query("xhtml:desired",xhtmlContext);
				if ( variants.size() == 0 ){
					out.println( "missing desired variants!");
					return;
				}
				Boolean result = false;
				StringBuffer iPortLink = new StringBuffer();
				String remark = "";
				try {
					for( int j=0; j < variants.size() && !result; ++j ){
						iPortLink.setLength(0);
						if ( variants.get(j).getChildCount() > 0 ){
							Node el = variants.get(j).getChild(0);
							result = testQuery( db, query, pos, el, iPortLink );
						}
						else
							result = testQuery( db, query, pos, null, iPortLink );
					}
				} catch ( Exception e) {
					remark = e.getMessage();
				}
				if ( remark.equals("") ){
					Nodes remNode = nodes.get(i).query("xhtml:remark",xhtmlContext);
					if ( remNode.size() > 0 ){
						if( remNode.get(0).getChildCount() > 0 )
							remark = remNode.get(0).getChild(0).toXML();
					}
				}
				if ( result ){
					logger.debug( "test " + (i+1) + " succeeded" );
					if ( !remark.equals("") )
						out.println("<tr bgcolor=\"orange\">");
					else
						out.println("<tr bgcolor=\"green\">");
					out.println( "<th>" + (i+1) + "</th><th>" + ln + "</th><th>" + db + "</th><th>" + "<a href=\"" + iPortLink.toString() + "\" target=\"_blank\">" + query + "</a></th><th>"  );
					if ( !remark.equals("") )
						out.println( "To Do");
					else
						out.println( "OK");
					out.println("</th>");
				}
				else {
					logger.debug( "test " + (i+1) + " failed" );
					out.println("<tr bgcolor=\"red\">");
					out.println( "<th>" + (i+1) + "</th><th>" + ln + "</th><th>" + db + "</th><th>" + "<a href=\"" + iPortLink.toString() + "\" target=\"_blank\">" + query + "</a></th><th>"  );
					out.println( "FAILED"  );
					out.println("</th>");
				}
				out.println( "<th>" + remark + "</th></tr>");
			}
			logger.debug( "done");
		}
		else
			out.println("no testCases found");
		out.println("</table>");
		out.println("</body>");
	}

	private String makeQuery(String select) {
		String query;
		if ( select == null ){
			query = "//xhtml:test";
			logger.debug("search for all test nodes" );
		}
		else {
			query = "//xhtml:test[xhtml:"+select+"!='']";
			logger.debug("search for all test nodes which contain " + select );
		}
		return query;
	}

	private String makeQuery(String tagName, String tagList, String select) {
		String[] parts = tagList.split( ",");
		if ( select == null )
			logger.debug("creating query for " + tagList + " nodes" );
		else
			logger.debug("creating query for " + tagList + " nodes whicht contain " + select );
		String query =  "//xhtml:test[(xhtml:" + tagName + "='" + parts[0] + "'";
		if ( parts.length > 1 ){
			for ( Integer k=1; k < parts.length; ++k)
				query += " or xhtml:" + tagName + "='" + parts[k] + "'";
		}
		query += ")";
		if ( select != null ){
			query += " and xhtml:"+select+"!=''";
		}
		query += "]";
		return query;
	}

	void doTestDb( PrintStream out, HttpServletRequest request ){
		String testDb = request.getParameter( "db" ); 	 
		String testId = request.getParameter( "id" ); 	 
		String select = request.getParameter( "select" );
		Document testDoc = getTestDoc( testerConfig.getProperty("testCases"));
		if ( testDoc != null ){
//			logger.debug("found testdoc:");
//			logger.debug(testDoc.toXML());
			Nodes nodes = null;
			if ( testDb != null ){
				String query = makeQuery( "database", testDb, select );
				logger.debug("Xpath query="+query);
				nodes = testDoc.query( query, xhtmlContext );
				if ( nodes.size() == 0 ){
					out.println( "no tests for database " + testDb + " found." );
					return;
				}
				logger.debug("found " + nodes.size() + " tests for " + testDb );
			}
			else {
				String query = makeQuery( select );
				nodes = testDoc.query( query, xhtmlContext );
				if ( nodes.size() == 0 ){
					out.println( "no tests found." );
					return;
				}
				logger.debug("found " + nodes.size() + " tests" );
			}
			Integer startIndex = 0;
			Integer endIndex = 0;
			if ( testId == null ){
				startIndex = 1;
				endIndex = nodes.size();
			}
			else {
//				logger.debug("try to find indexes in: " + testId);
				String[] parts = testId.split("-");
				if ( parts.length == 0 || parts.length > 2 ){
					out.println( "invalid index value: " + testId );
					return;
				}
//				for ( int i =0; i < parts.length; ++i ){
//				logger.debug( "part["+i+"]="+ parts[i] );
//				}
				startIndex = Integer.parseInt( parts[0] );
				if ( parts.length == 1 ){
					endIndex = startIndex;
				}
				else {
					endIndex = Integer.parseInt( parts[1]);
					if ( endIndex > nodes.size())
						endIndex=nodes.size();
				}
				if ( startIndex <= 0 || 
						startIndex > nodes.size() || 
						endIndex <= 0 ||
						endIndex < startIndex ) {
					out.println( "index out of range: " + testId );
					return;
				}
			}
			String dbCache = "";
			logger.debug( "Testing range " + startIndex + "-" + endIndex );
			for( int i=startIndex-1; i < endIndex; ++i){
				String db = nodes.get(i).query("xhtml:database", xhtmlContext).get(0).getValue().toString();
				if ( !db.equals( dbCache ) ){
					if ( !dbCache.equals("") )
						out.println("</table>");
					dbCache = db;
					out.println("<h3>Tests for " + db + "</h3>");
					out.println("<table border=\"1\">");
				}
				String query = nodes.get(i).query("xhtml:query",xhtmlContext).get(0).getValue().toString();
				Nodes indexNodes = nodes.get(i).query("xhtml:index",xhtmlContext);
				String index = "0";
				if ( indexNodes.size() > 0 )
					index = indexNodes.get(0).getValue().toString();
				String pos = "1";
				Nodes posNode = nodes.get(i).query("xhtml:position",xhtmlContext);
				if ( posNode.size() > 0 )
					pos = posNode.get(0).getValue().toString();
				Nodes variants = nodes.get(i).query("xhtml:desired",xhtmlContext);
				if ( variants.size() == 0 ){
					out.println( "missing desired variants!");
					return;
				}
				Boolean result = false;
				StringBuffer iPortLink = new StringBuffer();
				String remark = "";
				try {
					for( int j=0; j < variants.size() && !result; ++j ){
						iPortLink.setLength(0);
						if ( variants.get(j).getChildCount() > 0 ){
							Node el = variants.get(j).getChild(0);
							result = testQuery( db, query, pos, el, iPortLink );
						}
						else
							result = testQuery( db, query, pos, null, iPortLink );
					}
				}
				catch ( Exception e ){
					remark = e.getMessage();
				}
				if ( remark.equals("") ){
					Nodes remNode = nodes.get(i).query("xhtml:remark",xhtmlContext);
					if ( remNode.size() > 0 ){
						if( remNode.get(0).getChildCount() > 0 )
							remark = remNode.get(0).getChild(0).toXML();
					}
				}
				if ( result ){
					logger.debug( "test " + (i+1) + " succeeded" );
					if ( !remark.equals("") )
						out.println("<tr bgcolor=\"orange\">");
					else
						out.println("<tr bgcolor=\"green\">");
					out.println( "<th>" + index + "</th><th>" + "<a href=\"" + iPortLink.toString() + "\" target=\"_blank\">" + query + "</a></th><th>"  );
					if ( !remark.equals("") )
						out.println( "To Do");
					else
						out.println( "OK");
					out.println("</th>");
				}
				else {
					logger.debug( "test " + (i+1) + " failed" );
					out.println("<tr bgcolor=\"red\">");
					out.println( "<th>" + index + "</th><th>" + "<a href=\"" + iPortLink.toString() + "\" target=\"_blank\">" + query + "</a></th><th>"  );
					out.println( "FAILED"  );
					out.println("</th>");
				}
				out.println( "<th>" + remark + "</th></tr>");
			}
			logger.debug( "done");
		}
		else
			out.println("no testCases found");
		out.println("</table>");
		out.println("</body>");
	}

	private Document getTestDoc( String docName ) {
		Document result = null;
		if ( docName != null )
			try {
				logger.debug( "try to get XML file: " + docName );
				File f = getFile( docName );
				if ( f != null ){
					XMLReader tagsoup = XMLReaderFactory.createXMLReader("org.ccil.cowan.tagsoup.Parser");
					Builder builder = new Builder( tagsoup );
					try {
						result = builder.build( f );
					} 
					catch ( Exception e) {
						logger.error( "Tester:", e );
					}	
				}
			} catch (SAXException e ) {
				logger.debug("tagsoup creation problem", e );
			}
			return result;
	}

	private Boolean testQuery( String db, String query, String pos, Node testElement,
			StringBuffer iPortLinkBuf 	) throws Exception {
			String baseUrl= testerConfig.getProperty("iPortUrl");
			Properties args = new Properties();
			args.put("db", db);
			args.put("request", "full_record");
			args.put("position", pos );
			args.put("query", query );
			String iPortLink = Locator.createUrlString( baseUrl, args );
			logger.debug( "calling iPort: " + iPortLink );
			iPortLinkBuf.append(iPortLink);
			Document iPortResult = httpGetHTMLDocument( iPortLink, httpConfig );	
			if ( iPortResult != null){
//				logger.debug("GOT an iPort result!"); 
//				logger.debug( iPortResult.toXML() );
				Nodes nodes = iPortResult.query( "//xhtml:table/xhtml:tr[xhtml:td[starts-with(.,'LOCATOR')]]/xhtml:td[2]", xhtmlContext );
				if ( nodes.size() == 0 ){
					logger.debug( "No Locator result in document" );
//				logger.debug( iPortResult.toXML() );					
					if ( testElement == null )
						return true;
				}
				else {
					if ( equalSubNodes( nodes.get(0), testElement ) )
						return true;
					else {
						logger.debug( "comparing failed for: ");
						String cmp = nodes.get(0).toXML();
						String ok = "<>";
						if ( testElement != null )
							ok = testElement.toXML();
						logger.debug( "iport :" + cmp );
						logger.debug( "Wanted:" + ok );
						return false;
					}
				}
			}
		return false;
	}

	private boolean equalSubNodes(Node node, Node testElement) {
//		logger.debug( "equal subnodes ");
//		logger.debug( node.toXML() );
//		logger.debug( testElement.toXML() );
		if ( testElement != null ){
			// we only take <a> tags, except when there are none
			Nodes nl1 = node.query( "xhtml:a", xhtmlContext );
			Nodes nl2 = testElement.query( "xhtml:a", xhtmlContext );
			if ( nl1.size() == nl2.size() ){
				if ( nl1.size() == 0 ){
					String s1 = node.toXML();
					String s2 = testElement.toXML();
					s1 = s1.replaceAll("sessionid=\\d+", "" );
					s2 = s2.replaceAll("sessionid=\\d+", "" );
					return s1.equals(s2);
				}
				else {
					for ( int i = 0; i < nl1.size(); ++i ){
						Node n1 = nl1.get(i);
						boolean found = false;
						for ( int j=0; !found && j < nl2.size(); ++j ){
							Node n2 = nl2.get(j);
							String s1 = n1.toXML();
							String s2 = n2.toXML();
							if ( s1.indexOf("Full text via ILL request") > 0 ){
								// don't try to compare all evil details of an ILL 
								if ( s2.indexOf("Full text via ILL request") > 0 )
									found = true;
								else
									found = false;
							}
							else {
								s1 = s1.replaceAll("sessionid=\\d+", "" );
								s2 = s2.replaceAll("sessionid=\\d+", "" );
								if ( s1.equals(s2) ){
									found = true;
							  }
							}
						}
						if ( !found )
							return false;	
					}	
					return true;
				}	
			}
			else {
				logger.debug( "verschillend aantal <a> tags: " + nl1.size() + " vs. " + nl2.size() );	
			}
		}
		return false;
	}

	private nu.xom.Document httpGetHTMLDocument( String url, Properties httpProps )
	throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException {
		nu.xom.Document responseDocument = null;
		InputStream is = Locator.HttpGetInputStream(url, httpProps);
		try {
			XMLReader tagsoup = XMLReaderFactory.createXMLReader("org.ccil.cowan.tagsoup.Parser");
			Builder builder = new Builder( tagsoup );
			responseDocument =  builder.build(is);
		} catch (SAXException e) {
			logger.debug("tagsoup creation failed", e);
			e.printStackTrace();
		}
		finally {
			is.close();
		}
		return responseDocument;
	}

	static final long serialVersionUID=1;
}
