package nl.uvt.locator;

import java.util.LinkedList;
import java.util.List;

import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.XPathContext;

public class Threshold {
	static enum thKind {unspec, nested, year, volume, issue};
	static enum thCondition { errorC, lt, le, ge, gt };
	static enum act { fail, pass, checkmore };

	int value;
	thKind kind = thKind.unspec;
	thCondition condition = thCondition.errorC;
	List<Threshold> more = new LinkedList<Threshold>();

	act compareYear( int year ) throws LocatorException {
		if ( kind != thKind.year ){
			if ( kind == thKind.unspec && more != null ){
				Locator.logger.debug( "recurse " );
				act result = act.pass;
				for ( int i=0; i < more.size() && result==act.pass; ++i ){
					result = more.get(i).compareYear( year );
				}
				Locator.logger.debug( "recurse result=" + result.toString() );
				return result;
			}
			else
				return act.pass;
		}
		else {
			Locator.logger.debug( "Compare Year: " + year + " " + condition.toString() + " " + value );
			act result = act.fail;
			switch ( condition ){
			case lt:
				if ( year < value )
					result = act.pass;
				break;
			case le:
				if ( year < value )
					result = act.pass;
				else if ( year == value )
					result = act.checkmore;
				break;
			case ge:
				if ( year > value )
					result = act.pass;
				else if ( year == value )
					result = act.checkmore;
				break;
			case gt:
				if ( year > value )
					result = act.pass;
				break;
			default:
				throw new LocatorException( "invalid threshold year condition" );
			}	
			if ( result == act.fail )
				Locator.logger.debug( "Threshold year rejected: " + year 
						+ " " + condition.toString() + " " + value );
			else if ( result == act.checkmore )
				Locator.logger.debug( "Threshold year need more checking... " + year );
			return result;
		}
	}

	act compareVolume( int volume, int issue ) throws LocatorException {
		act result = act.pass;
		if ( kind == thKind.volume ){
			Locator.logger.debug( "Compare Volume: " + volume + " " + condition.toString() + " " + value );
			if ( volume < 0 )
				return act.fail;
			switch ( condition ){
			case lt:
				if ( volume >= value )
					result = act.fail;
				break;
			case le:
				if ( volume == value )
					result = act.checkmore;
				else if ( volume > value )
					result = act.fail;
				break;
			case ge:
				if ( volume == value )
					result = act.checkmore;
				else if ( volume < value )
					result = act.fail;
				break;
			case gt:
				if ( volume < value )
					result = act.fail;
				break;
			default:
				throw new LocatorException( "invalid threshold volume condition" );
			}	
		}
		if ( result == act.fail )
			Locator.logger.debug( "Threshold volume rejected; " + volume 
					+ " " + condition.toString() + " " + value );
		else if ( more != null && issue != 0 && result == act.checkmore ){
			// when we are at the boundaries of a volume, check issue also
			if ( issue < 0 )
				return act.pass;
			for ( int i=0; i < more.size() && result==act.checkmore; ++i ){
				result = more.get(i).compareIssue( issue );
			}
		}
		return result;
	}

	act compareIssue( int issue ) throws LocatorException {
		if ( kind != thKind.issue )
			return act.checkmore;
		else {
			Locator.logger.debug( "Compare Issue:" + issue + " " + condition.toString() + " " + value );
			act result = act.fail;
			switch ( condition ){
			case lt:
				if ( issue < value )
					result = act.pass;
				break;
			case le:
				if ( issue <= value )
					result = act.pass;
				break;
			case ge:
				if ( issue >= value )
					result = act.pass;
				break;
			case gt:
				if( issue > value )
					result = act.pass;
				break;
			default:
				throw new LocatorException( "invalid threshold condition 3" );
			}	
			if ( result == act.fail )
				Locator.logger.debug( "Threshold issue rejected: " + issue + 
						" " + condition.toString() + " " + value );
			return result;
		}
	}


	public static boolean inRange( List<Threshold> thresholds, QueryData pQ) throws LocatorException{
	act test = act.pass;	
	if ( pQ.year > 0 ){
		for ( int i=0; i < thresholds.size(); ++i ){
			test = thresholds.get(i).compareYear( pQ.year );
			if ( test==act.fail && thresholds.get(i).kind!=thKind.unspec )
				break;
		}
	}
	if ( test != act.fail ){
		if ( pQ.volume > 0 ){
			for ( int i=0; i < thresholds.size() && test != act.fail ; ++i ){
				test = thresholds.get(i).compareVolume( pQ.volume, pQ.issue );
			}
		}
		else {
			test= act.pass;
		}
	}
	Locator.logger.debug( "Threshold final result: " + test.toString());
	return test==act.pass;
}

	protected String getValue( Node node, String val, XPathContext context){
		Nodes nodes = node.query( val, context);
		if ( nodes.size() > 0 ){
			return nodes.get(0).getValue();
		}
		else
			return "";
	}

	protected String getValue( Node node, String val ){
		return getValue( node, val, null );
	}

	void fill( Node rec ) {
		String val = "";
		try {
			String attribute = getValue( rec, "attribute ");
			String operator = getValue( rec, "operator" );
			val = getValue( rec, "value" );
			if ( attribute.equals("date") ) { 
				kind = thKind.year;
			}
			if ( attribute.equals("volume") ) { 
				kind = thKind.volume;
			}
			if ( attribute.equals("issue") ) { 
				kind = thKind.issue;
			}
			if ( operator.equals("ge") ){
				condition = thCondition.ge;
				value = Integer.parseInt(val.trim());
			}
			else if ( operator.equals("gt") ){
				condition = thCondition.gt;
				value = Integer.parseInt(val.trim()) ;
			}
			else if ( operator.equals("le") ){
				condition = thCondition.le;
				value = Integer.parseInt( val.trim());
			}
			else if ( operator.equals("lt") ){
				condition = thCondition.lt;
				value = Integer.parseInt( val.trim() );
			}
		}
		catch ( NumberFormatException e ){
			// quite a hack
			// some issues are entered like 2-3. Maybe Volumes too?
			// anyway: take the 'correct' value
			String[] parts = val.split("-");
			if ( parts.length == 2 ){
				if ( condition == thCondition.lt )
					value = Integer.parseInt( parts[0].trim() );
				else if ( condition == thCondition.le )
					value = Integer.parseInt( parts[1].trim() );
				else if ( condition == thCondition.gt )
					value = Integer.parseInt( parts[1].trim() );
				else if ( condition == thCondition.ge )
					value = Integer.parseInt( parts[0].trim() );
			}
			else {
				Locator.logger.warn("invalid values in treshold:" + val );
			}
		}
		Nodes nodes = rec.query( "threshold ");
		for ( int i=0; i < nodes.size(); ++i ){
			Threshold th = new Threshold();
			th.fill( nodes.get(i) );
			more.add( th );
		}
	}

	void fill( Node rec, String ns, XPathContext context ) {
		String val = "";
		try {
//			Locator.logger.debug("fill from: " + rec.toXML() );
			String attribute = getValue( rec, ns+":attribute", context);
			String operator = getValue( rec, ns+":operator", context );
			val = getValue( rec, ns+":value", context );
			if ( attribute.equals( "date") ) { 
				kind = thKind.year;
			}
			if ( attribute.equals("volume") ) { 
				kind = thKind.volume;
			}
			if ( attribute.equals("issue") ) { 
				kind = thKind.issue;
			}
			if ( operator.equals("ge") ){
				condition = thCondition.ge;
				value = Integer.parseInt(val);
			}
			else if ( operator.equals("gt") ){
				condition = thCondition.gt;
				value = Integer.parseInt(val) ;
			}
			else if ( operator.equals("le") ){
				condition = thCondition.le;
				value = Integer.parseInt( val);
			}
			else if ( operator.equals("lt") ){
				condition = thCondition.lt;
				value = Integer.parseInt( val );
			}
		}
		catch ( NumberFormatException e ){
			// quite a hack
			// some issues are entered like 2-3. Maybe Volumes too?
			// anyway: take the 'correct' value
			String[] parts = val.split("-");
			if ( parts.length == 2 ){
				if ( condition == thCondition.lt )
					value = Integer.parseInt( parts[0] );
				else if ( condition == thCondition.le )
					value = Integer.parseInt( parts[1] );
				else if ( condition == thCondition.gt )
					value = Integer.parseInt( parts[1] );
				else if ( condition == thCondition.ge )
					value = Integer.parseInt( parts[0] );
			}
			else {
				Locator.logger.warn("invalid values in treshold:" + val );
			}
		}
		Nodes nodes = rec.query( ns+":threshold", context );
		for ( int i=0; i < nodes.size(); ++i ){
			Threshold th = new Threshold();
			th.fill( nodes.get(i) );
			more.add( th );
		}
	}

	public String toString(){
		return toString( false );
	}

	public String toString( boolean nested ){
		String result = "";
		if ( condition == thCondition.lt ){
			if ( !nested )
				result += "-"; 
			else {
				switch ( kind ){
				case  year: 
					result += "year:";
					break;
				case volume:
					result += "volume:";
					break;
				case issue:
					result += "issue:";
					break;
				}
			}
			result += value - 1;
		}
		else if ( condition == thCondition.le ){
			if ( !nested )
				result += "-"; 
			else {
				switch ( kind ){
				case  year: 
					result += "year:";
					break;
				case volume:
					result += "volume:";
					break;
				case issue:
					result += "issue:";
					break;
				}
			}
			result += value;
		}
		else if ( condition == thCondition.ge ){
			switch ( kind ){
			case  year: 
				result += " year:";
				break;
			case volume:
				result += " volume:";
				break;
			case issue:
				result += " issue:";
				break;
			}
			result += value;
		}
		else if ( condition ==thCondition.gt){
			switch ( kind ){
			case  year: 
				result += " year:";
				break;
			case volume:
				result += " volume:";
				break;
			case issue:
				result += " issue:";
				break;
			}
			result += value;
		}

		if ( more != null )
			for ( int i=0; i < more.size(); ++i ){
				result += more.get(i).toString( true );
			}
		return result;
	}
}
